// Node.cpp


#include "stdafx.h"
#include "Node.h"

#include "SimuRoute.h"
#include "GraphView.h"
#include "Doc.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


IMPLEMENT_SERIAL(CNode, CObject, VERSIONABLE_SCHEMA | 1)


// Constructor
CNode::CNode()
{
	SetType(nodeInvalid);
	m_pt.SetPoint(-1,-1);
	m_iOrient = 0;
	m_len     = 0;
	
	m_nd      = app.m_ndDefault;
	
	m_sel     = selNo;
}


// Affectation operator
CNode& CNode::operator = (const CNode& node)
{
	SetType(node.m_type);
	m_pt      = node.m_pt;
	m_iOrient = node.m_iOrient;
	m_len     = node.m_len;
	
	m_nd      = node.m_nd;
	
	m_sel     = selNo;
	
	return *this;
}


// Set the node type
void CNode::SetType(NODETYPE type)
{
	m_type = type;
	
	switch (m_type) {
		
		case nodeT:
			m_nbLink = 3;
			break;
		
		case nodeX:
			m_nbLink = 4;
			break;
		
		default:
			m_nbLink = 0;
			break;
	}
}


// Serialization
void CNode::Serialize(CArchive& ar)
{
	if (ar.IsStoring()) {
		// Saving
		
		ar << (int)m_type;
		ar << m_pt << m_iOrient << m_len;
		ar.Write(&m_nd, sizeof(m_nd));
		
	}else{
		// Loading
		
		int type;
		ar >> type;
		SetType((NODETYPE)type);
		
		ar >> m_pt >> m_iOrient >> m_len;
		switch (ar.GetObjectSchema()) {
			
			case 0:
				ar.Read(&m_nd, sizeof(NODEDATA0));
				m_nd.m_fStats = true;
				break;
			
			// Last format
			default:
				ar.Read(&m_nd, sizeof(m_nd));
				break;
		}
	}
}


// Get the size in cells
CSize CNode::GetCellsSize(bool fStrict) const
{
	ASSERT(m_type!=nodeInvalid);
	
	CSize size(2,2);
	
	switch (m_type) {
		
		// End of road
		// Corner
		case nodeEnd:
		case nodeCorner:
			size.cx = size.cy = 0;
			break;
		
		// T node
		case nodeT:
			if (fStrict) {
				if (m_iOrient & 1)
					size.cy--;
				else
					size.cx--;
			}
			break;
		
		// Line
		case nodeLine:
			size     = aoffDir[m_iOrient];
			size.cx *= m_len;
			size.cy *= m_len;
			break;
	}
	
	return size;
}


// Get the rectangle in cells
CRect CNode::GetCellsRect(bool fStrict) const
{
	ASSERT(m_type!=nodeInvalid);
	
	CPoint pt = m_pt;
	
	if (m_type==nodeT && fStrict) {
		// T strict
		
		if (m_iOrient==0)
			pt.x++;
		else if (m_iOrient==1)
			pt.y++;
	}
	
	CRect rc(pt, GetCellsSize(fStrict));
	rc.NormalizeRect();
	rc.right++;
	rc.bottom++;
	return rc;
}


// Get the rectangle in pixels
void CNode::GetRect(const CGraphView* pView, CRect& rc, bool fStrict) const
{
	ASSERT(m_type!=nodeInvalid);
	
	rc = GetCellsRect(fStrict);
	pView->Convert(rc);
}


// Turn the node
void CNode::Turn(CDoc* pDoc, int iOffset, bool fMustBeValid)
{
	ASSERT(!fMustBeValid || pDoc->IsOverlapOK(*this));
	
	// Turn while there is a problem
	do
		m_iOrient = RoundOrient(m_iOrient + iOffset);
	while (fMustBeValid && !pDoc->IsOverlapOK(*this));
}


// Draw the node (on pView->m_aCellShadow)
void CNode::Draw(const CGraphView* pView, CDC& dc, bool fSimul) const
{
	ASSERT(m_type!=nodeInvalid);
	
	CRect rc = GetCellsRect();
	rc &= pView->GetDocument()->GetBoardRect();
	const CSize size = pView->GetDocument()->m_size;
	
	for (int y=rc.top; y<rc.bottom; y++)
		ZeroMemory(pView->m_aCellShadow + rc.left + y*size.cx,
			rc.Width()*sizeof(pView->m_aCellShadow[0]));
	
	UpdateCells(pView->m_aCellShadow, size);
	pView->DrawCells(dc, pView->m_aCellShadow, rc, fSimul);
}


// Update m_iOrient with a offset
// (the longest dimension determines the orientation)
void CNode::SetOrient(CSize sizeOrient)
{
	if (m_type==nodeCorner) {
		// Use the diagonals
		
		m_iOrient =
			(sizeOrient.cx < 0)
				? ((sizeOrient.cy < 0) ? 3 : 2)
				: ((sizeOrient.cy < 0) ? 0 : 1);
		
	}else{
		// Use the axis
		
		m_iOrient = (abs(sizeOrient.cx) >= abs(sizeOrient.cy))
			? ((sizeOrient.cx < 0) ? 3 : 1)
			: ((sizeOrient.cy < 0) ? 0 : 2);
	}
}

// Update m_len and m_iOrient with m_pt and pt2
// sizeOrient is used to have a orientation if m_pt==pt2
void CNode::SetPoint2(CPoint pt2, CSize sizeOrient)
{
	const CSize size = pt2 - m_pt;
	
	if (size.cx || size.cy)
		sizeOrient = size;
	
	m_len = max(abs(size.cx), abs(size.cy));
	SetOrient(sizeOrient);
}


static const int aWallEnd[4] =
{
	maskWalls & ~maskT | maskLimitT,
	maskWalls & ~maskR | maskLimitR,
	maskWalls & ~maskB | maskLimitB,
	maskWalls & ~maskL | maskLimitL,
};

static const int aWallCorner[4] =
{
	maskB | maskL | maskLimitT | maskLimitR | maskTR,
	maskT | maskL | maskLimitB | maskLimitR | maskBR,
	maskT | maskR | maskLimitB | maskLimitL | maskBL,
	maskB | maskR | maskLimitT | maskLimitL | maskTL,
};

static const int aWallT[4][3*3] =
{
	{
		maskNothing, maskL | maskR | maskLimitT, 0,
		maskNothing, maskL | maskTR | maskBR, maskT | maskB | maskLimitR,
		maskNothing, maskL | maskR | maskLimitB, 0,
	},
	{
		maskNothing, maskNothing, maskNothing,
		maskT | maskB | maskLimitL, maskT | maskBL | maskBR, maskT | maskB | maskLimitR,
		0, maskL | maskR | maskLimitB, 0,
	},
	{
		0, maskL | maskR | maskLimitT, maskNothing,
		maskT | maskB | maskLimitL, maskR | maskTL | maskBL, maskNothing,
		0, maskL | maskR | maskLimitB, maskNothing,
	},
	{
		0, maskL | maskR | maskLimitT, 0,
		maskT | maskB | maskLimitL, maskB | maskTL | maskTR, maskT | maskB | maskLimitR,
		maskNothing, maskNothing, maskNothing,
	},
};

static const int aWallX[3*3] =
{
	0, maskL | maskR | maskLimitT, 0,
	maskT | maskB | maskLimitL, maskPoints, maskT | maskB | maskLimitR,
	0, maskL | maskR | maskLimitB, 0,
};


// Update the array "cell position -> node"
void CNode::UpdateCells(CELL aCell[], CSize size) const
{
	CRect rc = GetCellsRect(false);
	
	const int *aWall = NULL;
	int nb = 0;
	
	switch (m_type) {
		
		// End of road
		case nodeEnd:
			aWall = &aWallEnd[m_iOrient];
			nb    = 1;
			break;
		
			// Corner
		case nodeCorner:
			aWall = &aWallCorner[m_iOrient];
			nb    = 1;
			break;
		
		// T node
		case nodeT:
			aWall = aWallT[m_iOrient];
			nb    = nbArray(aWallT[m_iOrient]);
			break;
		
		// X node
		case nodeX:
			aWall = aWallX;
			nb    = nbArray(aWallX);
			break;
		
		// Line
		case nodeLine:
			if (IsHorz())
				for (int x=rc.left; x<rc.right; x++)
					ModifyCell(aCell, size, x,m_pt.y, maskT | maskB);
			else
				for (int y=rc.top; y<rc.bottom; y++)
					ModifyCell(aCell, size, m_pt.x,y, maskL | maskR);
			return;
	}
	
	
	// Fill the CELL structures
	for (; rc.top<rc.bottom; rc.top++)
		for (int x=rc.left; x<rc.right; x++)
			ModifyCell(aCell, size, x,rc.top, *aWall++);
}


// Modify a cell in an array
void CNode::ModifyCell(CELL aCell[], CSize size, int x, int y, int iWall) const
{
	if (x<0 || y<0 || x>=size.cx || y>=size.cy)
		return;
	if (iWall==maskNothing)
		return;
	
	if (iWall)
		iWall |= maskRoad;
	
	// Modify the cell
	CELL &cell = aCell[x + y*size.cx];
	ASSERT(!cell.pNode);
	cell.pNode = (CNode*)this;
	cell.iWall = iWall;
}



// Update the UNIT array of a view with the node
void CNode::UpdateUnits(CGraphView* pView) const
{
	CSize asizeDiagDir[4];
	for (int iDiagDir=0; iDiagDir<4; iDiagDir++)
		asizeDiagDir[iDiagDir] = aoffDiagDir[RoundOrient(m_iOrient+iDiagDir)];
	
	CPoint ptUnit = pView->GetToUnit(m_pt);
	
	const CPoint ptCenter(m_pt.x+1, m_pt.y+1);
	
	
	switch (m_type) {
		
		// End of road
		case nodeEnd:
			{
				// One generator
				GENERATOR &gen  = pView->m_lstGen.GetAt(
					pView->m_lstGen.AddTail(GENERATOR()));
				ZeroMemory(&gen, sizeof(gen));
				gen.pNode       = this;
				gen.pt          = ptUnit+asizeDiagDir[0];
				gen.probaNewCar = m_nd.m_probaNewCar;
				gen.dir         = m_iOrient;
				
				// One output (for the generator)
				pView->SetUnit(gen.pt, m_nd.m_maxSpeed, NULL,
					m_iOrient, 1);
				
				// On input: a "pit" road to remove cars
				pView->SetUnit(ptUnit+asizeDiagDir[3], m_nd.m_maxSpeed, NULL);
				
				for (int iDir=0; iDir<4; iDir++)
					pView->GetUnit(ptUnit+aoffDiagDir[iDir]).pGen = &gen;
			}
			break;
		
		
		// Corner
		case nodeCorner:
			{
				static const int aOrient[4] = { 0, 1, 1, 2 };
				
				for (int iDiagDir=0; iDiagDir<4; iDiagDir++)
					pView->SetUnit(ptUnit+asizeDiagDir[iDiagDir], m_nd.m_maxSpeed, NULL,
						RoundOrient(m_iOrient + aOrient[iDiagDir]), 1);
			}
			break;
		
		
		// T node
		case nodeT:
			{
				const int iOrient1 = RoundOrient(m_iOrient + 1);
				const int iOrient2 = RoundOrient(m_iOrient + 2);
				
				const CPoint ptOrient0 = pView->GetToUnit(ptCenter, m_iOrient);
				const CPoint ptOrient1 = pView->GetToUnit(ptCenter,   iOrient1);
				
				
				// Create the two lights
				CLight &light1 = pView->AddLight(m_nd,
					ptOrient0 + aoffDir[m_iOrient+4],
					iOrient2,
					true);
				CLight &light2 = pView->AddLight(m_nd,
					ptOrient1 + aoffDir[iOrient1+4],
					RoundOrient(m_iOrient + 3),
					false);
				
				// - "Legs" of the node (a cell each time, 3 in the whole node)
				// - Center of the node (a cell)
				for (int iDir=0; iDir<4; iDir++) {
					if (iDir!=RoundOrient(m_iOrient+3))
						pView->SetFourUnits(ptCenter+aoffDir[iDir], iDir,
							m_nd.m_maxSpeed,
							(iDir==RoundOrient(m_iOrient+1)) ? &light1 : &light2);
				}
				
				
				pView->SetUnit(ptOrient0,
					m_nd.m_maxSpeed, NULL,
					m_iOrient,    m_nd.m_aProba[m_iOrient],
					iOrient2 + 4, m_nd.m_aProba[iOrient2]);
				pView->SetUnit(ptOrient1,
					m_nd.m_maxSpeed, NULL,
					m_iOrient, m_nd.m_aProba[m_iOrient],
					iOrient1,  m_nd.m_aProba[iOrient1]);
				pView->SetUnit(pView->GetToUnit(ptCenter, m_iOrient+2),
					m_nd.m_maxSpeed, NULL,
					iOrient2, m_nd.m_aProba[iOrient2]);
				pView->SetUnit(pView->GetToUnit(ptCenter, m_iOrient+3),
					m_nd.m_maxSpeed, NULL,
					iOrient2,     m_nd.m_aProba[iOrient2],
					iOrient1 + 4, m_nd.m_aProba[iOrient1]);
			}
			break;
		
		
		// X node
		case nodeX:
			
			// - "Legs" of the cross (a cell each time, 4 in the whole node)
			// - Center of the cross (a cell)
			for (int iDir=0; iDir<4; iDir++) {
				const CPoint ptUnit = pView->GetToUnit(ptCenter, iDir);
				const int iDir1 = RoundOrient(iDir + 1);
				const int iDir2 = RoundOrient(iDir + 2);
				const int iDir3 = RoundOrient(iDir + 3);
				
				// Create the light
				CLight &light = pView->AddLight(m_nd,
					pView->GetToUnit(ptCenter, iDir1) + aoffDir[iDir1+4],
					iDir3,
					ToBool(iDir & 1));
				
				pView->SetFourUnits(ptCenter + aoffDir[iDir], iDir,
					m_nd.m_maxSpeed, &light);
				
				pView->SetUnit(ptUnit, m_nd.m_maxSpeed, NULL,
					iDir,      m_nd.m_aProba[iDir],
					iDir2 + 4, m_nd.m_aProba[iDir2],
					iDir3,     m_nd.m_aProba[iDir3]);
				
				// Delete a possible move
				MOVE_DIR &md = pView->GetUnit(ptUnit).aMoveDir[iDir];
				md.nbMove--;
			}
			break;
		
		
		// Line
		case nodeLine:
			{
				CPoint ptCell = m_pt;
				for (int len=0; len<=m_len; len++, ptCell+=aoffDir[m_iOrient])
					pView->SetFourUnits(ptCell, m_iOrient, m_nd.m_maxSpeed, NULL);
			}
			break;
	}
}
